import logging
import os
from typing import Any, Optional, Union

import molflux
from datasets import Dataset

try:
    import lightning.pytorch as pl
    import torch

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning_gp", e) from e

from molflux.modelzoo.errors import NotTrainedError
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.model import StandardDeviationMixin
from molflux.modelzoo.models.lightning.config import (
    CompileConfig,
    DataModuleConfig,
    OptimizerConfig,
    SchedulerConfig,
    TrainerConfig,
    TransferLearningConfigBase,
)
from molflux.modelzoo.models.lightning.model import LightningModelBase
from molflux.modelzoo.models.lightning_gp.gp_config import GPConfig
from molflux.modelzoo.models.lightning_gp.gp_datamodule import GPDataModule
from molflux.modelzoo.models.lightning_gp.gp_module import GPModule
from molflux.modelzoo.typing import PredictionResult

logger = logging.getLogger(__name__)


_DESCRIPTION = """
Implementation of Gaussian Processes via Pytorch Lightning and GPyTorch.

A GP in a nutshell, is a model that takes in as training inputs all your training data.
It then passes the data through a series of mathematical operations that depend on a mean function,
a kernel (how close two data points are) and optionally any pre or post processing objects
(a feature extractor before, an added noise after etc.).

The output of this is a multivariate gaussian, for which you can compute a loss if you have the labels for the
training data. Note that the mean, kernel, added noise, feature extractors etc. can have learnable parameters, and
the whole process is differentiable, so we can (and do) do gradient descent to find the best fit for these parameters.

GPyTorch allows users to quickly define a torch compatible GP, but don't offer out of the box any training loops.
functionality in a nice set of configs, motivating combining the two. The PR implements Exact GPs, that require the
entire training set, which means the memory complexity is ~O(N^2)log N, but offers the most precise models when the
data range is within the memory limit.
"""

_CONFIG_DESCRIPTION = """
Parameters
----------
iterations : Optional[int], default: None
"""


class LightningGPRegressor(StandardDeviationMixin, LightningModelBase[GPConfig]):
    """Model class for MLP model"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description=_DESCRIPTION,
            config_description=_CONFIG_DESCRIPTION,
        )

    @property
    def _datamodule_builder(self) -> type[GPDataModule]:
        return GPDataModule

    @property
    def _config_builder(self) -> type[GPConfig]:
        return GPConfig

    def _instantiate_module(self) -> GPModule:
        return GPModule(model_config=self.model_config)

    def _predict(
        self,
        data: Dataset,
        datamodule_config: DataModuleConfig | dict[str, Any] | None = None,
        trainer_config: TrainerConfig | dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> PredictionResult:
        return self._predict_with_std(
            data=data,
            datamodule_config=datamodule_config,
            trainer_config=trainer_config,
            **kwargs,
        )[0]

    def _predict_with_std(
        self,
        data: Dataset,
        datamodule_config: DataModuleConfig | dict[str, Any] | None = None,
        trainer_config: TrainerConfig | dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[PredictionResult, PredictionResult]:
        """Implements a single prediction with standard deviation step."""
        del kwargs

        # assume single task models
        mean_display_names, std_display_names = list(
            self._predict_with_std_display_names,
        )

        # if data is empty
        if not len(data):
            return {mean_display_names[0]: []}, {std_display_names[0]: []}

        with self.override_config(datamodule=datamodule_config, trainer=trainer_config):
            datamodule = self._instantiate_datamodule(predict_data=data)

            trainer_config = self.model_config.trainer.pass_to_trainer()
            trainer = pl.Trainer(**trainer_config)

            # Expect a list of Tensors, which may need to be overwritten for some Torch models
            batch_preds: list[tuple[torch.Tensor]] = trainer.predict(
                self.module,
                datamodule,
            )

        # assume a single batch with tuples
        means, stds = batch_preds[0][0].reshape(-1), batch_preds[0][1].reshape(-1)  # type: ignore[misc]

        output = {mean_display_names[0]: means.tolist()}
        output_std = {std_display_names[0]: stds.tolist()}

        return output, output_std

    def as_dir(self, directory: str) -> None:
        """Saves the model into a directory."""

        if hasattr(self, "module") and self.module is not None:
            ckpt = {
                "state_dict": self.state_dict,
                "train_inputs": self.module.gp_model.train_inputs,
                "train_targets": self.module.gp_model.train_targets,
            }
            torch.save(ckpt, os.path.join(directory, "module_checkpoint.ckpt"))
        else:
            raise NotTrainedError("No initialised module to save.")

    def from_dir(self, directory: str) -> None:
        """Loads a model from a directory."""

        ckpt_path = os.path.join(directory, "module_checkpoint.ckpt")
        checkpoint = torch.load(ckpt_path, map_location="cpu", weights_only=True)

        self.module = self._instantiate_module()
        self.state_dict = checkpoint["state_dict"]
        self.module.gp_model.set_train_data(
            inputs=checkpoint["train_inputs"],
            targets=checkpoint["train_targets"],
            strict=False,
        )

    def _train_multi_data(
        self,
        train_data: dict[Optional[str], Dataset],
        validation_data: Union[
            dict[Optional[str], Dataset],
            None,
        ] = None,
        datamodule_config: Union[DataModuleConfig, dict[str, Any], None] = None,
        trainer_config: Union[TrainerConfig, dict[str, Any], None] = None,
        optimizer_config: Union[OptimizerConfig, dict[str, Any], None] = None,
        scheduler_config: Union[SchedulerConfig, dict[str, Any], None] = None,
        transfer_learning_config: Union[
            TransferLearningConfigBase,
            dict[str, Any],
            None,
        ] = None,
        compile_config: Union[CompileConfig, dict[str, Any], bool, None] = None,
        ckpt_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        if self.model_config.validation_config.name == "InnerTrainValidation":
            logger.info("Doing internal validation split")
            assert validation_data is None, RuntimeError(
                "Can't do internal validation split if validation_data provided",
            )

            strategy = molflux.splits.load_from_dict(
                self.model_config.validation_config.splitting_strategy_config,
            )
            split_datasets: dict[str, dict[Any, Any]] = {
                "train": {},
                "validation": {},
            }
            for k, v in train_data.items():
                split_data = next(
                    molflux.datasets.split_dataset(
                        v,
                        strategy=strategy,
                        groups_column=self.model_config.validation_config.groups_column,
                        target_column=self.model_config.validation_config.target_column,
                    ),
                )

                assert len(split_data["train"]) > 0, RuntimeError(
                    "Train split resulted in 0 points in train split",
                )
                assert len(split_data["validation"]) > 0, RuntimeError(
                    "Validation split resulted in 0 points in validation split",
                )
                assert len(split_data["test"]) == 0, RuntimeError(
                    "Test split resulted in non-zero points in test split",
                )

                logger.info(
                    f"{k} validation split: train ({len(split_data['train'])}), validation ({len(split_data['validation'])})",
                )
                split_datasets["train"][k] = split_data["train"]
                split_datasets["validation"][k] = split_data["validation"]

            super()._train_multi_data(
                train_data=split_datasets["train"],
                validation_data=split_datasets["validation"],
                datamodule_config=datamodule_config,
                trainer_config=trainer_config,
                optimizer_config=optimizer_config,
                scheduler_config=scheduler_config,
                transfer_learning_config=transfer_learning_config,
                compile_config=compile_config,
                ckpt_path=ckpt_path,
                **kwargs,
            )
        else:
            super()._train_multi_data(
                train_data=train_data,
                validation_data=validation_data,
                datamodule_config=datamodule_config,
                trainer_config=trainer_config,
                optimizer_config=optimizer_config,
                scheduler_config=scheduler_config,
                transfer_learning_config=transfer_learning_config,
                compile_config=compile_config,
                ckpt_path=ckpt_path,
                **kwargs,
            )



from typing import Any, Literal, Union

from pydantic.v1 import Field, dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict


@dataclasses.dataclass(config=ConfigDict)
class BackboneConfig:
    name: str


@dataclasses.dataclass(config=ConfigDict)
class NoBackboneConfig(BackboneConfig):
    name: Literal["NoBackbone"] = "NoBackbone"


@dataclasses.dataclass(config=ConfigDict)
class LayerConfig:
    name: str
    parameters: dict[str, Any] = Field(default_factory=dict)


@dataclasses.dataclass(config=ConfigDict)
class NNBackboneConfig(BackboneConfig):
    name: Literal["NNBackbone"] = "NNBackbone"
    layer_configs: list[LayerConfig] = Field(default_factory=list)


BackboneConfigT = Union[
    NoBackboneConfig,
    NNBackboneConfig,
]



from pydantic.v1 import Field, dataclasses, root_validator

from molflux.modelzoo.models.lightning.config import (
    ConfigDict,
    LightningConfig,
)
from molflux.modelzoo.models.lightning_gp.configs.backbone import (
    BackboneConfigT,
    NoBackboneConfig,
)
from molflux.modelzoo.models.lightning_gp.configs.kernel import (
    KernelConfigT,
    RBFKernelConfig,
)
from molflux.modelzoo.models.lightning_gp.configs.likelihood import LikelihoodConfig
from molflux.modelzoo.models.lightning_gp.configs.mean import MeanConfig
from molflux.modelzoo.models.lightning_gp.configs.validation import (
    InnerTrainValidationConfig,
    ValidationConfigT,
)


@dataclasses.dataclass(config=ConfigDict)
class GPConfig(LightningConfig):
    mean_config: MeanConfig = Field(default_factory=MeanConfig)
    likelihood_config: LikelihoodConfig = Field(default_factory=LikelihoodConfig)
    kernel_config: KernelConfigT = Field(
        default_factory=RBFKernelConfig,
        discriminator="name",
    )
    num_tasks: int = 1
    validation_config: ValidationConfigT = Field(
        default_factory=InnerTrainValidationConfig,
        discriminator="name",
    )
    backbone_config: BackboneConfigT = Field(
        default_factory=NoBackboneConfig,
        discriminator="name",
    )

    @root_validator()
    def check_single_task(cls, values: dict) -> dict:
        assert values["num_tasks"] == 1, ValueError(
            "lightning_gp_regressor only supports single task models!",
        )
        return values

    @root_validator(pre=False, skip_on_failure=True)
    def disable_sanity_val_steps(cls, values: dict) -> dict:
        trainer = values["trainer"]
        trainer.num_sanity_val_steps = 0
        return values



import pandas as pd
import pytest

import datasets
from molflux.modelzoo.load import load_from_dict
from molflux.modelzoo.protocols import Model

model_name = "lightning_gp_regressor"

_X_FEATURES = ["X_col_1", "X_col_2"]
_Y_FEATURES = ["y_col"]


@pytest.fixture()
def train_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            [1.0, 2.0, 3.0],
            [1.0, 3.0, 4.0],
            [2.0, 4.0, 6.0],
            [2.0, 4.0, 6.0],
            [2.0, 4.0, 6.0],
        ],
        columns=_X_FEATURES + _Y_FEATURES,
    )


@pytest.fixture()
def validation_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            [1.0, 2.0, 3.0],
            [1.0, 3.0, 4.0],
            [2.0, 4.0, 6.0],
            [2.0, 4.0, 6.0],
            [2.0, 5.0, 7.0],
        ],
        columns=_X_FEATURES + _Y_FEATURES,
    )


@pytest.fixture()
def predict_df() -> pd.DataFrame:
    return pd.DataFrame(
        [
            [1.0, 2.0],
            [1.0, 3.0],
            [2.0, 4.0],
            [2.0, 4.0],
            [2.0, 5.0],
        ],
        columns=_X_FEATURES,
    )


@pytest.fixture()
def empty_predict_df() -> pd.DataFrame:
    return pd.DataFrame([], columns=_X_FEATURES)


@pytest.fixture()
def train_dataset(train_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(train_df)


@pytest.fixture()
def train_dataset_b(train_df: pd.DataFrame) -> datasets.Dataset:
    joint_df = pd.concat([train_df, train_df]).iloc[:8]
    return datasets.Dataset.from_pandas(joint_df)


@pytest.fixture()
def validation_dataset(validation_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(validation_df)


@pytest.fixture()
def predict_dataset(predict_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(predict_df)


@pytest.fixture()
def empty_predict_dataset(empty_predict_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(empty_predict_df)


@pytest.fixture()
def single_row_predict_dataset(predict_df: pd.DataFrame) -> datasets.Dataset:
    return datasets.Dataset.from_pandas(predict_df.iloc[:1])


@pytest.fixture(scope="function")
def fixture_gp_config() -> dict:
    gp_config = {
        "name": model_name,
        "config": {
            "x_features": _X_FEATURES,
            "y_features": _Y_FEATURES,
            "num_tasks": len(_Y_FEATURES),
            "likelihood_config": {
                "noise_constraint": {
                    "name": "GreaterThan",
                    "lower_bound": 0.01,
                },
            },
            "kernel_config": {
                "name": "RBFKernel",
            },
            "trainer": {
                "accelerator": "cpu",
            },
        },
    }
    return gp_config


@pytest.fixture(scope="function")
def fixture_gp_model(fixture_gp_config: dict) -> Model:
    return load_from_dict(fixture_gp_config)



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.openeye.descriptors.molecular_weight import (
    MolecularWeight,
)

representation_name = "molecular_weight"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, MolecularWeight)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_compute(fixture_representation):
    """That default scoring gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples)
    expected_result = [458.4147]
    assert representation_name in result
    assert result[representation_name] == pytest.approx(expected_result)


def test_batch_compute(fixture_representation):
    """That batch scoring gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
        "CC",
    ]
    result = representation.featurise(samples)
    expected_result = [458.4147, 30.069]
    assert representation_name in result
    assert result[representation_name] == pytest.approx(expected_result, abs=0.0005)


def test_default_digitise(fixture_representation):
    """Test default digitised featurisation gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples, digitise=True)
    expected_result = [[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0]]
    assert representation_name in result
    assert result[representation_name] == expected_result



from pydantic.v1 import Field, dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict
from molflux.modelzoo.models.lightning_gp.configs.constraint import (
    ConstraintConfigT,
    PositiveConstraintConfig,
)


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class LikelihoodConfig:
    noise_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )



try:
    import gpytorch
    import torch

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning_gp", e) from e

from molflux.modelzoo.models.lightning_gp.configs.utils import (
    generate_kernel,
    generate_mean,
)
from molflux.modelzoo.models.lightning_gp.gp_config import GPConfig


class GPModel(gpytorch.models.ExactGP):
    def __init__(
        self,
        model_config: GPConfig,
        train_x: torch.Tensor,
        train_y: torch.Tensor,
        likelihood: gpytorch.likelihoods.Likelihood,
    ):
        super().__init__(train_x, train_y, likelihood)
        self.model_config = model_config

        self.mean_module = generate_mean(model_config.mean_config)
        self.covar_module = generate_kernel(model_config.kernel_config)

    def forward(self, x: torch.Tensor) -> gpytorch.distributions.MultivariateNormal:
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)



from typing import Literal, Union

from pydantic.v1 import dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class IntervalConstraintConfig:
    name: Literal["Interval"] = "Interval"
    lower_bound: float
    upper_bound: float


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class PositiveConstraintConfig:
    name: Literal["Positive"] = "Positive"


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class GreaterThanConstraintConfig:
    lower_bound: float
    name: Literal["GreaterThan"] = "GreaterThan"


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class LessThanConstraintConfig:
    upper_bound: float
    name: Literal["LessThan"] = "LessThan"


ConstraintConfigT = Union[
    IntervalConstraintConfig,
    PositiveConstraintConfig,
    GreaterThanConstraintConfig,
    LessThanConstraintConfig,
]



from typing import cast

import gpytorch
import torch
from gpytorch.constraints import GreaterThan, Interval, LessThan, Positive
from gpytorch.kernels import (
    AdditiveKernel,
    CosineKernel,
    LinearKernel,
    MaternKernel,
    PeriodicKernel,
    PiecewisePolynomialKernel,
    PolynomialKernel,
    ProductKernel,
    RBFKernel,
    RQKernel,
    ScaleKernel,
)

from molflux.modelzoo.models.lightning_gp.configs.backbone import (
    BackboneConfigT,
    NNBackboneConfig,
)
from molflux.modelzoo.models.lightning_gp.configs.constraint import (
    ConstraintConfigT,
    GreaterThanConstraintConfig,
    IntervalConstraintConfig,
    LessThanConstraintConfig,
)
from molflux.modelzoo.models.lightning_gp.configs.kernel import (
    AdditiveKernelConfig,
    CosineKernelConfig,
    JaccardKernelConfig,
    KernelConfig,
    LinearKernelConfig,
    MaternKernelConfig,
    PeriodicKernelConfig,
    PiecewisePolynomialKernelConfig,
    PolynomialKernelConfig,
    ProductKernelConfig,
    RBFKernelConfig,
    RQKernelConfig,
)
from molflux.modelzoo.models.lightning_gp.configs.mean import MeanConfig
from molflux.modelzoo.models.lightning_gp.gp_config import GPConfig
from molflux.modelzoo.models.lightning_gp.kernels.jaccard import JaccardKernel


def generate_backbone(backbone_config: BackboneConfigT) -> torch.nn.Module | None:
    name = backbone_config.name
    if name == "NoBackbone":
        return None

    if name == "NNBackbone":
        backbone_config = cast(NNBackboneConfig, backbone_config)

        layers = []
        for layer_config in backbone_config.layer_configs:
            layer_name = layer_config.name
            layer_parameters = layer_config.parameters
            layer_class = getattr(torch.nn, layer_name)
            layer = layer_class(**layer_parameters)
            layers.append(layer)

        return torch.nn.Sequential(*layers)

    else:
        raise RuntimeError(f"Backbone config name {name} not recognised!")


def generate_mean(mean_config: MeanConfig) -> gpytorch.means.Mean:
    fixed = mean_config.fixed
    constant_value = mean_config.constant_value

    mean = gpytorch.means.ConstantMean()
    if constant_value is not None:
        mean.constant.data = torch.tensor(constant_value, dtype=torch.float)
    mean.constant.requires_grad = not fixed

    return mean


def generate_constraint(constraint_config: ConstraintConfigT | None) -> Interval | None:
    if constraint_config is None:
        return None

    if constraint_config.name == "Interval":
        constraint_config = cast(IntervalConstraintConfig, constraint_config)  # type: ignore[redundant-cast]
        constraint = Interval(
            lower_bound=constraint_config.lower_bound,
            upper_bound=constraint_config.upper_bound,
        )

    elif constraint_config.name == "Positive":
        constraint = Positive()

    elif constraint_config.name == "GreaterThan":
        constraint_config = cast(GreaterThanConstraintConfig, constraint_config)  # type: ignore[redundant-cast]
        constraint = GreaterThan(lower_bound=constraint_config.lower_bound)

    elif constraint_config.name == "LessThan":
        constraint_config = cast(LessThanConstraintConfig, constraint_config)  # type: ignore[redundant-cast]
        constraint = LessThan(upper_bound=constraint_config.upper_bound)

    else:
        raise ValueError(
            f"Unrecognised constraint config name - {constraint_config.name}",
        )

    return constraint


def generate_kernel(kernel_config: KernelConfig) -> gpytorch.kernels.Kernel:
    name = kernel_config.name

    if name == "CosineKernel":
        kernel_config = cast(CosineKernelConfig, kernel_config)
        period_length_constraint = generate_constraint(
            kernel_config.period_length_constraint,
        )
        kernel = CosineKernel(
            period_length_constraint=period_length_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "JaccardKernel":
        kernel_config = cast(JaccardKernelConfig, kernel_config)
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        kernel = JaccardKernel(
            lengthscale_constraint=lengthscale_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "LinearKernel":
        kernel_config = cast(LinearKernelConfig, kernel_config)
        variance_constraint = generate_constraint(kernel_config.variance_constraint)
        kernel = LinearKernel(
            variance_constraint=variance_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "MaternKernel":
        kernel_config = cast(MaternKernelConfig, kernel_config)
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        kernel = MaternKernel(
            nu=kernel_config.nu,
            lengthscale_constraint=lengthscale_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "PeriodicKernel":
        kernel_config = cast(PeriodicKernelConfig, kernel_config)
        period_length_constraint = generate_constraint(
            kernel_config.period_length_constraint,
        )
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        kernel = PeriodicKernel(
            period_length_constraint=period_length_constraint,
            lengthscale_constraint=lengthscale_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "PiecewisePolynomicalKernel":
        kernel_config = cast(PiecewisePolynomialKernelConfig, kernel_config)
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        kernel = PiecewisePolynomialKernel(
            q=kernel_config.q,
            lengthscale_constraint=lengthscale_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "PolynomicalKernel":
        kernel_config = cast(PolynomialKernelConfig, kernel_config)
        offset_constraint = generate_constraint(kernel_config.offset_constraint)
        kernel = PolynomialKernel(
            power=kernel_config.power,
            offset_constraint=offset_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "RBFKernel":
        kernel_config = cast(RBFKernelConfig, kernel_config)
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        kernel = RBFKernel(
            lengthscale_constraint=lengthscale_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "RQKernel":
        kernel_config = cast(RQKernelConfig, kernel_config)
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        alpha_constraint = generate_constraint(kernel_config.alpha_constraint)
        kernel = RQKernel(
            lengthscale_constraint=lengthscale_constraint,
            alpha_constraint=alpha_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "JaccardKernel":
        kernel_config = cast(JaccardKernelConfig, kernel_config)
        lengthscale_constraint = generate_constraint(
            kernel_config.lengthscale_constraint,
        )
        kernel = JaccardKernel(
            lengthscale_constraint=lengthscale_constraint,
            ard_num_dims=kernel_config.ard_num_dims,
            active_dims=kernel_config.active_dims,
            eps=kernel_config.eps,
        )
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "AdditiveKernel":
        kernel_config = cast(AdditiveKernelConfig, kernel_config)

        kernels = [generate_kernel(config) for config in kernel_config.kernels]

        kernel = AdditiveKernel(*kernels)
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    elif name == "ProductKernel":
        kernel_config = cast(ProductKernelConfig, kernel_config)

        kernels = [generate_kernel(config) for config in kernel_config.kernels]

        kernel = ProductKernel(*kernels)
        if kernel_config.add_scale:
            kernel = ScaleKernel(kernel)

    else:
        raise ValueError(f"Kernel with {name=} not supported!")

    return kernel


def generate_likelihood(model_config: GPConfig) -> gpytorch.likelihoods.Likelihood:
    noise_constraint_config = model_config.likelihood_config.noise_constraint
    constraint = generate_constraint(noise_constraint_config)

    return gpytorch.likelihoods.GaussianLikelihood(noise_constraint=constraint)



from typing import Any, cast

import datasets
from molflux.modelzoo.models.lightning.datamodule import LightningDataModule
from molflux.modelzoo.models.lightning_gp.gp_config import GPConfig

try:
    import torch
    from torch.utils.data import Dataset

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning_gp", e) from e


class GPDataset(Dataset):
    def __init__(
        self,
        dataset: datasets.Dataset,
        x_features: list[str],
        y_features: list[str],
        with_y_features: bool,
    ) -> None:
        self.dataset = dataset
        self.x_features = x_features
        self.y_features = y_features
        self.with_y_features = with_y_features

        all_features = self.x_features
        if with_y_features:
            all_features = all_features + self.y_features

        self.dataset.set_format("torch", columns=all_features)

    def __len__(self) -> int:
        # we send the whole dataset through, in a single batch
        return 1

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        # assume we're doing single batch optimisation
        del idx

        x_list: list[torch.Tensor] = []
        for x_col in self.x_features:
            x_feat = cast(torch.Tensor, self.dataset[x_col])

            if len(x_feat.shape) == 1:
                x_feat = x_feat.unsqueeze(1)

            x_list.append(x_feat)

        x = torch.cat(x_list, dim=-1)

        if self.with_y_features:
            y_list: list[torch.Tensor] = []
            for y_col in self.y_features:
                y_feat = cast(torch.Tensor, self.dataset[y_col])

                y_list.append(y_feat)

            y = torch.cat(y_list, dim=0)
        else:
            y = torch.empty(0)

        return x, y


class GPDataModule(LightningDataModule):
    model_config: GPConfig

    def __init__(
        self,
        model_config: GPConfig,
        train_data: dict[str | None, datasets.Dataset] | None = None,
        validation_data: dict[str | None, datasets.Dataset] | None = None,
        test_data: dict[str | None, datasets.Dataset] | None = None,
        predict_data: datasets.Dataset | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            model_config=model_config,
            train_data=train_data,
            validation_data=validation_data,
            test_data=test_data,
            predict_data=predict_data,
            **kwargs,
        )

    def prepare_dataset(
        self,
        data: datasets.Dataset,
        split: str,
        name: str | None = None,
        **kwargs: Any,
    ) -> Dataset:
        del name, kwargs

        with_y_features = split != "predict"

        return GPDataset(
            dataset=data,
            x_features=self.model_config.x_features,
            y_features=self.model_config.y_features,
            with_y_features=with_y_features,
        )



import logging
from typing import Any, Optional

try:
    import gpytorch
    import gpytorch.mlls.marginal_log_likelihood
    import torch

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning_gp", e) from e

from molflux.modelzoo.models.lightning.module import (
    LightningModuleBase,
    SingleBatchStepOutput,
)
from molflux.modelzoo.models.lightning_gp.configs.utils import (
    generate_backbone,
    generate_likelihood,
)
from molflux.modelzoo.models.lightning_gp.gp_config import GPConfig
from molflux.modelzoo.models.lightning_gp.gpytorch_model import GPModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class GPModule(LightningModuleBase):
    model_config: GPConfig

    def __init__(self, model_config: GPConfig) -> None:
        super().__init__(model_config=model_config)

        self.likelihood = generate_likelihood(self.model_config)
        self.gp_model = GPModel(
            model_config=model_config,
            train_x=torch.empty(0).float(),
            train_y=torch.empty(0).float(),
            likelihood=self.likelihood,
        )

        self.backbone = generate_backbone(model_config.backbone_config)

        self.loss = gpytorch.mlls.ExactMarginalLogLikelihood(
            self.likelihood,
            self.gp_model,
        )
        self.validation_data: Optional[tuple[torch.Tensor, torch.Tensor]] = None
        self.train_data_set = False

    def forward(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        # if we have a backbone, pass the X features (first arg) through the backbone, and then to the GP
        if self.backbone is not None:
            features = self.backbone(args[0], **kwargs)
            return self.gp_model(features, *(args[1:]), **kwargs)

        else:
            return self.gp_model(*args, **kwargs)

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        """fit method for doing a predict step"""
        del batch_idx, dataloader_idx

        x, _ = batch
        x = x.squeeze(
            0,
        )  # the BxNxD tensor is squeezed as we always have a batch of size 1

        outputs = self.likelihood(self(x))

        return outputs.loc, outputs.stddev

    def _training_step_on_single_source_batch(
        self,
        single_source_batch: Any,
        source_name: str | None,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> SingleBatchStepOutput:
        """Implements the training step on one minibatch.

        Returns a dict with a loss key and a batch size."""
        del source_name, batch_idx, args, kwargs

        x, y = single_source_batch
        x, y = x.squeeze(0), y.squeeze(0)

        if self.gp_model.train_inputs[0].shape[0] != x.shape[0]:
            logger.info(
                f"Set the train data for the GP, from {self.gp_model.train_inputs[0].shape[0]} "
                f"to {x.shape[0]} data points",
            )
            self.gp_model.set_train_data(inputs=x, targets=y, strict=False)
            self.train_data_set = True

        outputs = self(x)
        loss = -self.loss(outputs, target=y)

        batch_size = x.shape[0]

        return loss, {}, batch_size

    def _validation_step_on_single_source_batch(
        self,
        single_source_batch: Any,
        source_name: str | None,
        batch_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> SingleBatchStepOutput:
        del source_name, batch_idx, args, kwargs

        x, y = single_source_batch
        x, y = x.squeeze(0), y.squeeze(0)
        self.validation_data = x, y

        outputs = self.likelihood(self(x))
        loss = gpytorch.metrics.negative_log_predictive_density(outputs, y)

        batch_size = x.shape[0]

        return loss, {}, batch_size

    def on_fit_end(self) -> None:
        """Add the validation data to the model at the end of training, if specified"""
        if (
            self.model_config.validation_config.fit_end_fantasy
            and self.validation_data is not None
        ):
            self.gp_model.set_train_data(
                inputs=torch.cat(
                    [self.gp_model.train_inputs[0], self.validation_data[0]],
                ),
                targets=torch.cat(
                    [self.gp_model.train_targets, self.validation_data[1]],
                ),
                strict=False,
            )



from typing import Any

try:
    import torch
    from gpytorch.kernels import Kernel

except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning_gp", e) from e


class JaccardKernel(Kernel):
    """
    Computes a covariance matrix based on the Jaccard kernel between inputs.
    Assumes that the input tensor is a binary bit vector.
    """

    has_lengthscale = True

    def forward(
        self,
        x1: torch.Tensor,
        x2: torch.Tensor,
        diag: bool = False,
        last_dim_is_batch: bool = False,
        **params: Any,
    ) -> torch.Tensor:
        x1_ = x1
        x2_ = x2

        # if the input is 2D, make it 3D. This is done for allowing broadcasting
        # for the divisor computation
        if x1_.dim() == 2:
            x1_ = x1_.unsqueeze(0)
            x2_ = x2_.unsqueeze(0)
            must_squeeze = True
        else:
            must_squeeze = False

        if last_dim_is_batch:
            x1_ = x1_.transpose(-1, -2).unsqueeze(-1)
            x2_ = x2_.transpose(-1, -2).unsqueeze(-1)

        prod = torch.matmul(x1_, x2_.transpose(-2, -1))

        divisor = (
            torch.sum(x1_, dim=-1).unsqueeze(-1)
            + torch.sum(x2_, dim=-1).unsqueeze(1)
            - prod
        )

        result: torch.Tensor = (prod / divisor) * self.lengthscale
        if must_squeeze:
            result = result.squeeze(0)
        if diag:
            result = result.diag()

        return result



from pydantic.v1 import dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class MeanConfig:
    fixed: bool = False
    constant_value: float | None = None



import numpy as np
import torch

from molflux.modelzoo import load_from_dict, load_from_store, save_to_store
from molflux.modelzoo.catalogue import list_models
from molflux.modelzoo.models.lightning_gp.gp_model import LightningGPRegressor
from molflux.modelzoo.protocols import Estimator, SupportsStandardDeviation

model_name = "lightning_gp_regressor"


def test_in_catalogue():
    """That the model is registered in the catalogue."""
    catalogue = list_models()
    all_names = [name for names in catalogue.values() for name in names]
    assert model_name in all_names


def test_default_model_tag_matches_entrypoint_name(fixture_gp_model):
    """That the default model tag matches the catalogue entrypoint name.

    This is not strictly required, but ensures a more intuitive user experience.
    """
    model = fixture_gp_model
    assert model.tag == model_name


def test_is_mapped_to_correct_class(fixture_gp_model):
    """That the model name is mapped to the appropriate class."""
    model = fixture_gp_model
    assert isinstance(model, LightningGPRegressor)


def test_implements_protocol(fixture_gp_model):
    """That the model implements the public Estimator protocol."""
    model = fixture_gp_model
    assert isinstance(model, Estimator)


def test_implements_standard_deviation_protocol(fixture_gp_model):
    """That the model implements the standard deviation protocol."""
    model = fixture_gp_model
    assert isinstance(model, SupportsStandardDeviation)


def test_train_predict_model(
    fixture_gp_model,
    tmp_path,
    train_dataset,
    predict_dataset,
    empty_predict_dataset,
    single_row_predict_dataset,
):
    """Test that a model can run the train, predict and predict_with_std functions"""

    model = fixture_gp_model

    model.train(
        train_data=train_dataset,
        validation_data=None,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 5,
            "default_root_dir": str(tmp_path),
        },
    )

    for prediction_dataset in [
        empty_predict_dataset,
        single_row_predict_dataset,
        predict_dataset,
    ]:
        predictions = model.predict(prediction_dataset)
        assert len(predictions) == len(model.y_features)

        for task in predictions:
            assert len(predictions[task]) == len(prediction_dataset)

        predictions_from_std_method, stds = model.predict_with_std(prediction_dataset)
        assert predictions_from_std_method and stds
        assert len(predictions_from_std_method) == len(model.y_features) and len(
            stds,
        ) == len(model.y_features)
        for task in predictions_from_std_method:
            assert len(predictions_from_std_method[task]) == len(predictions[task])


def test_saving_loading(
    tmp_path,
    fixture_gp_model,
    train_dataset,
    predict_dataset,
):
    """Test that a model can successfully be saved and loaded"""

    fixture_gp_model.train(
        train_data=train_dataset,
        validation_data=None,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 5,
            "default_root_dir": str(tmp_path),
        },
    )

    save_to_store(tmp_path, fixture_gp_model)
    loaded_model = load_from_store(tmp_path)

    assert loaded_model.predict(predict_dataset)


def test_train_model_overfit(
    fixture_gp_model,
    tmp_path,
    train_dataset,
):
    """Test that a model can overfit to a small batch"""

    torch.manual_seed(0)

    model = fixture_gp_model

    model.train(
        train_data=train_dataset,
        validation_data=None,
        trainer_config={
            "accelerator": "cpu",
            "max_epochs": 400,
            "default_root_dir": str(tmp_path),
        },
        optimizer_config={"name": "SGD", "config": {"lr": 1}},
    )
    predictions = model.predict(
        train_dataset,
    )

    np.testing.assert_allclose(
        predictions[f"{model_name}::y_col"],
        train_dataset["y_col"],
        rtol=5e-2,
    )


def test_validation_configs(fixture_gp_config, train_dataset, predict_dataset):
    """Test that running with the NoValidation parameter works"""
    config = fixture_gp_config
    config["config"]["validation_config"] = {"name": "NoValidation"}

    model = load_from_dict(config)
    model.train(train_dataset)
    predictions = model.predict(train_dataset)

    assert len(predictions) == len(model.y_features)



from typing import Literal, Union

from pydantic.v1 import Field, dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class ValidationConfig:
    name: str
    fit_end_fantasy: bool = True


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class NoValidationConfig(ValidationConfig):
    name: Literal["NoValidation"] = "NoValidation"


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class InnerTrainValidationConfig(ValidationConfig):
    name: Literal["InnerTrainValidation"] = "InnerTrainValidation"
    splitting_strategy_config: dict = Field(
        default_factory=lambda: {
            "name": "shuffle_split",
            "presets": {
                "train_fraction": 0.8,
                "validation_fraction": 0.2,
                "test_fraction": 0.0,
            },
        },
    )
    groups_column: str | None = None
    target_column: str | None = None


ValidationConfigT = Union[
    NoValidationConfig,
    InnerTrainValidationConfig,
]



from typing import Literal, Union

from pydantic.v1 import Field, dataclasses

from molflux.modelzoo.models.lightning.config import ConfigDict
from molflux.modelzoo.models.lightning_gp.configs.constraint import (
    ConstraintConfigT,
    PositiveConstraintConfig,
)


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class KernelConfig:
    name: str
    ard_num_dims: int | None = None
    lengthscale_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )
    eps: float = 1e-6
    active_dims: list[int] | None = None
    add_scale: bool = False


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class CosineKernelConfig(KernelConfig):
    name: Literal["CosineKernel"] = "CosineKernel"
    period_length_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class JaccardKernelConfig(KernelConfig):
    name: Literal["JaccardKernel"] = "JaccardKernel"


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class LinearKernelConfig(KernelConfig):
    name: Literal["LinearKernel"] = "LinearKernel"
    variance_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class MaternKernelConfig(KernelConfig):
    name: Literal["MaternKernel"] = "MaternKernel"
    nu: float = 2.5


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class PeriodicKernelConfig(KernelConfig):
    name: Literal["PeriodicKernel"] = "PeriodicKernel"
    period_length_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class PiecewisePolynomialKernelConfig(KernelConfig):
    name: Literal["PiecewisePolynomialKernel"] = "PiecewisePolynomialKernel"
    q: Literal[0, 1, 2, 3] = 2


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class PolynomialKernelConfig(KernelConfig):
    name: Literal["PolynomialKernel"] = "PolynomialKernel"
    offset_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )
    power: int = 2


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class RBFKernelConfig(KernelConfig):
    name: Literal["RBFKernel"] = "RBFKernel"


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class RQKernelConfig(KernelConfig):
    name: Literal["RQKernel"] = "RQKernel"
    alpha_constraint: ConstraintConfigT = Field(
        default_factory=PositiveConstraintConfig,
        discriminator="name",
    )


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class AdditiveKernelConfig(KernelConfig):
    kernels: list["KernelConfigT"]
    name: Literal["AdditiveKernel"] = "AdditiveKernel"


@dataclasses.dataclass(config=ConfigDict, kw_only=True)
class ProductKernelConfig(KernelConfig):
    kernels: list["KernelConfigT"]
    name: Literal["ProductKernel"] = "ProductKernel"


KernelConfigT = Union[
    CosineKernelConfig,
    JaccardKernelConfig,
    LinearKernelConfig,
    MaternKernelConfig,
    PeriodicKernelConfig,
    PiecewisePolynomialKernelConfig,
    PolynomialKernelConfig,
    RBFKernelConfig,
    RQKernelConfig,
    AdditiveKernelConfig,
    ProductKernelConfig,
]

AdditiveKernelConfig.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]
ProductKernelConfig.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]



import pytest

from molflux.features import Representation, list_representations, load_representation
from molflux.features.representations.openeye.fingerprints.circular import (
    Circular,
)

representation_name = "circular"


@pytest.fixture(scope="function")
def fixture_representation() -> Representation:
    return load_representation(representation_name)


def test_representation_in_catalogue():
    """That the representation is registered in the catalogue."""
    catalogue = list_representations()
    all_representation_names = [name for names in catalogue.values() for name in names]
    assert representation_name in all_representation_names


def test_representation_is_mapped_to_correct_class(fixture_representation):
    """That the catalogue name is mapped to the appropriate class."""
    representation = fixture_representation
    assert isinstance(representation, Circular)


def test_implements_protocol(fixture_representation):
    """That the representation implements the public Representation protocol."""
    representation = fixture_representation
    assert isinstance(representation, Representation)


def test_default_fingerprint_length(fixture_representation):
    """That default featurisation gives fingerptins of expected length."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples)
    assert representation_name in result
    fingerprint = result[representation_name][0]
    assert len(fingerprint) == 2048


def test_custom_fingerprint_length(fixture_representation):
    """That can calculate fingerprints of custom lengths."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    length = 64
    result = representation.featurise(samples, length=length)
    fingerprint = result[representation_name][0]
    assert len(fingerprint) == length


def test_empty_smiles_fingerprint_length(fixture_representation):
    """That empty SMILES strings are processed correctly and give fingerprints of the expected length."""
    representation = fixture_representation
    samples = [""]
    length = 16
    result = representation.featurise(samples, length=length)
    fingerprint = result[representation_name][0]
    assert len(fingerprint) == length


def test_all_fingerprints_have_same_length(fixture_representation):
    """That can calculate fingerprints of custom lengths.

    Also make sure that empty SMILES strings are handled correctly.
    """
    representation = fixture_representation
    samples = ["C", "", "c1ccccc1"]
    length = 32
    result = representation.featurise(samples, length=length)
    assert all(
        len(fingerprint) == length for fingerprint in result[representation_name]
    )


def test_basic_compute(fixture_representation):
    """That default (shortened) featurisation gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
    ]
    result = representation.featurise(samples, length=16)
    expected_result = [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0]]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_batch_compute(fixture_representation):
    """That batch scoring gives expected results."""
    representation = fixture_representation
    samples = [
        "Cc1ccc2c(=O)c3cccc(CC(=O)OC4OC(C(=O)O)[C@@H](O)[C@@H](O)[C@@H]4O)c3oc2c1C",
        "CC",
    ]
    result = representation.featurise(samples, length=16)
    expected_result = [
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0],
    ]
    assert representation_name in result
    assert result[representation_name] == expected_result


def test_representation_has_default_state(fixture_representation):
    assert fixture_representation.state != {}
    assert "samples" not in fixture_representation.state
    assert "kwargs" not in fixture_representation.state
    fixture_representation.update_state(length=4096, diameter=10)
    assert fixture_representation._default_state != fixture_representation.state
    assert fixture_representation.state["length"] == 4096
    assert fixture_representation.state["diameter"] == 10
